"""
The io module contains functions to read/send data and write results to 
files/html reports.
"""
import pandas as pd
import numpy as np
import logging
import os
from os.path import abspath, dirname, join
import pecos.graphics
import datetime
from jinja2 import Environment, PackageLoader
import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.application import MIMEApplication
import base64

try:
    from sqlalchemy import create_engine
    import minimalmodbus 
except:
    pass

try:
    from nose.tools import nottest as _nottest
except ImportError:
    def _nottest(afunction):
        return afunction
        
logger = logging.getLogger(__name__)

env = Environment(loader=PackageLoader('pecos', 'templates'))

def read_campbell_scientific(file_name, index_col='TIMESTAMP', encoding=None):
    """
    Read Campbell Scientific CSV file.

    Parameters
    ----------
    file_name : string
        File name, with full path

    index_col : string (optional)
        Index column name, default = 'TIMESTAMP'

    encoding : string (optional)
        Character encoding (i.e. utf-16)
    
    Returns
    ---------
    pandas DataFrame with data
    """
    logger.info("Reading Campbell Scientific CSV file " + file_name)

    try:
        df = pd.read_csv(file_name, skiprows=1, encoding=encoding, index_col=index_col, parse_dates=True, dtype ='unicode', error_bad_lines=False) #, low_memory=False)
        df = df[2:]
        index = pd.to_datetime(df.index)
        Unnamed = df.filter(regex='Unnamed')
        df = df.drop(Unnamed.columns, 1)
        df = pd.DataFrame(data = df.values, index = index, columns = df.columns, dtype='float64')
    except:
        logger.warning("Cannot extract database, CSV file reader failed " + file_name)
        df = pd.DataFrame()
        return

    # Drop rows with NaT (not a time) in the index
    try:
        df.drop(pd.NaT, inplace=True)
    except:
        pass

    return df
    
def send_email(subject, body, recipient, sender, attachment=None, 
               host='localhost', username=None, password=None):
    """
    Send email using Python smtplib and email packages.
    
    Parameters
    ----------
    subject : string
        Subject text
        
    body : string
        Email body, in HTML or plain format
    
    recipient : list of string
        Recipient email address or addresses
    
    sender : string
        Sender email address
        
    attachment : string (optional)
        Name of file to attach
        
    host : string (optional)
        Name of email host (or host:port), default = 'localhost'
    
    username : string (optional)
        Email username for authentication
    
    password : string (optional)
        Email password for authentication
    """
    
    logger.info("Sending email")
    
    msg = _create_email_message(subject, body, recipient, sender)
    
    if attachment is not None:
        fp = open(attachment, "rb")  # Read as a binary file, even if it's text  
        att = MIMEApplication(fp.read())
        att.add_header('Content-Disposition', 'attachment', 
                       filename=os.path.basename(attachment))
        fp.close()
        msg.attach(att)
    
    s = smtplib.SMTP(host)
    try: # Authentication
        s.ehlo()
        s.starttls()
        s.login(username, password)
    except:
        pass
    s.sendmail(sender, recipient, msg.as_string())
    s.quit()
    
def _create_email_message(subject, body, recipient, sender):
    
    msg = MIMEMultipart()
    msg['Subject'] = subject
    msg['To'] = ', '.join(recipient)
    msg['From'] = sender
    
    if "</html>" in body.lower():
        content = MIMEText(body, 'html')
    else:
        content = MIMEText(body, 'plain')
    
    msg.attach(content)
    
    return msg
        
def write_metrics(filename, metrics):
    """
    Write metrics file.
    
    Parameters
    -----------
    filename : string
        File name, with full path
    
    metrics : pandas DataFrame
        Data to add to the metrics file
    """
    logger.info("Write metrics file")

    try:
        previous_metrics = pd.read_csv(filename, index_col='TIMESTEP') #, parse_dates=True)
    except:
        previous_metrics = pd.DataFrame()
        
    metrics.index = metrics.index.to_native_types() # this is necessary when using time zones
    metrics = metrics.combine_first(previous_metrics) 
    
    fout = open(filename, 'w')
    metrics.to_csv(fout, index_label='TIMESTEP', na_rep = 'NaN')
    fout.close()

@_nottest
def write_test_results(filename, test_results):
    """
    Write test results file.

    Parameters
    -----------
    filename : string
        File name, with full path

    test_results : pandas DataFrame
        Test results stored in pm.test_results
    """

    test_results.sort_values(list(test_results.columns), inplace=True)
    test_results.index = np.arange(1, test_results.shape[0]+1)

    logger.info("Writing test results csv file " + filename)

    fout = open(filename, 'w')
    test_results.to_csv(fout, na_rep = 'NaN')
    fout.close()

def write_monitoring_report(filename, pm, test_results_graphics=[], custom_graphics=[], metrics=None, 
                            title='Pecos Monitoring Report', config={}, logo=False, 
                            im_width_test_results=700, im_width_custom=700, encode=False):
    """
    Generate a monitoring report.  
    The monitoring report is used to report quality control test results for a single system.
    The report includes custom graphics, performance metrics, and test results.
    
    Parameters
    ----------
    filename : string
        File name, with full path

    pm : pecos PerformanceMonitoring object
        Contains data (pm.df) and test results (pm.test_results)
    
    test_results_graphics : list of strings (optional)
        Graphics files, with full path.  These graphics highlight data points 
        that failed a quality control test, created using pecos.graphics.plot_test_results()
        
    custom_graphics : list of strings (optional)
        Custom files, with full path.  Created by the user.
    
    metrics : pandas DataFrame (optional)
        Performance metrics to add as a table to the monitoring report
    
    title : string (optional)
        Monitoring report title, default = 'Pecos Monitoring Report'
        
    config : dictionary (optional)
        Configuration options, to be printed at the end of the report
    
    logo : string (optional)
        Graphic to be added to the report header
    
    im_width_test_results=700 : float (optional)
        Image width for test results graphics in the HTML report, default = 700
    
    im_width_custom=700 : float (optional)
        Image width for custom graphics in the HTML report, default = 700
        
    encode : boolean (optional)
        Encode graphics in the html, default = False
    """
    
    logger.info("Writing HTML report")
    
    if pm.df.empty:
        logger.warning("Empty database")
        start_time = 'NaN'
        end_time = 'NaN'
    else:
        start_time = pm.df.index[0]
        end_time = pm.df.index[-1]
    
    # Set pandas display option     
    pd.set_option('display.max_colwidth', -1)
    pd.set_option('display.width', 40)
    
    # Collect notes (from the logger file)
    try:
        logfiledir = os.path.join(dirname(abspath(__file__)))
        f = open(join(logfiledir,'logfile'), 'r')
        notes = f.read()
        f.close()
        notes_df = pd.DataFrame(notes.splitlines())
        notes_df.index += 1
    except:
        notes_df = pd.DataFrame()
    
    pm.test_results.sort_values(list(pm.test_results.columns), inplace=True)
    pm.test_results.index = np.arange(1, pm.test_results.shape[0]+1)
    #pm.test_results.reset_index(inplace=True)
    
    # Convert to html format
    if metrics is None:
        metrics = pd.DataFrame()
    test_results_html = pm.test_results.to_html(justify='left')
    metrics_html = metrics.to_html(justify='left')
    notes_html = notes_df.to_html(justify='left', header=False)
    
    content = {'start_time': str(start_time), 
                'end_time': str(end_time), 
                'num_notes': str(notes_df.shape[0]),
                'notes': notes_html, 
                'num_test_results': str(pm.test_results.shape[0]),
                'test_results': test_results_html,
                'test_results_graphics': test_results_graphics,
                'custom_graphics': custom_graphics,
                'num_metrics': str(metrics.shape[0]),
                'metrics': metrics_html,
                'config': config}
                
    title = os.path.basename(title)
    
    html_string = _html_template_monitoring_report(content, title, logo, im_width_test_results, im_width_custom, encode)
    
    # Write html file
    html_file = open(filename,"w")
    html_file.write(html_string)
    html_file.close()
    
    logger.info("")
    
def write_dashboard(filename, column_names, row_names, content, 
                    title='Pecos Dashboard', footnote='', logo=False, im_width=250, datatables=False, encode=False):
    """
    Generate a dashboard.  
    The dashboard is used to compare results across multiple systems.
    Each cell in the dashboard includes custom system graphics and metrics.
    
    Parameters
    ----------
    filename : string
        File name, with full path
    
    column_names : list of strings
        Column names listed in the order they should appear in the dashboard, i.e. ['location1', 'location2']
        
    row_names : list of strings
        Row names listed in the order they should appear in the dashboard, i.e. ['system1', 'system2']
        
    content : dictionary
        Dashboard content for each cell. 
        
        Dictionary keys are tuples indicating the row name and column name, i.e. ('row name', 'column name'), where 'row name' is in the list row_names and 'column name' is in the list column_names. 
        
        For each key, another dictionary is defined that contains the content to be included in each cell of the dashboard.
        Each cell can contain text, graphics, a table, and an html link.  These are defined using the following keys:
        
        - text (string) =  text at the top of each cell
        - graphics (list of strings) =  a list of graphics file names.  Each file name includes the full path
        - table (string) = a table in html format, for example a table of performance metrics.  DataFrames can be converted to an html string using df.to_html() or df.transpose().to_html().  Values in the table can be color coded using pandas Styler class. 
        - link (dict) = a dictionary where keys define the name of the link and values define the html link (with full path)
        
        For example::
        
            content = {('row name', 'column name'): {
                'text': 'text at the top', 
                'graphic': ['C:\\\\pecos\\\\results\\\\custom_graphic.png'], 
                'table': df.to_html(), 
                'link': {'Link to monitoring report': 'C:\\\\pecos\\\\results\\\\monitoring_report.html'}}
        
    title : string (optional)
        Dashboard title, default = 'Pecos Dashboard'
    
    footnote : string (optional)
        Text to be added to the end of the report
    
    logo : string (optional)
        Graphic to be added to the report header
    
    im_width : float (optional)
        Image width in the HTML report, default = 250
        
    datatables : boolean (optional)
        Use datatables.net to format the dashboard, default = False.  See https://datatables.net/ for more information.
    
    encode : boolean (optional)
        Encode graphics in the html, default = False
    """
    
    logger.info("Writing dashboard")
        
    # Set pandas display option     
    pd.set_option('display.max_colwidth', -1)
    pd.set_option('display.width', 40)
    
    html_string = _html_template_dashboard(column_names, row_names, content, title, footnote, logo, im_width, datatables, encode)
    
    # Write html file
    html_file = open(filename,"w")
    html_file.write(html_string)
    html_file.close()
    
    logger.info("")

def _html_template_monitoring_report(content, title, logo, im_width_test_results, im_width_custom, encode):
    
    # if encode == True, encode the images
    img_dic = {}
    if encode:
        for im in content['custom_graphics']:
            img_encode = base64.b64encode(open(im, "rb").read()).decode("utf-8")
            img_dic[im] = img_encode
        for im in content['test_results_graphics']:
            img_encode = base64.b64encode(open(im, "rb").read()).decode("utf-8")
            img_dic[im] = img_encode

    template = env.get_template('monitoring_report.html')

    date = datetime.datetime.now()
    datestr = date.strftime('%m/%d/%Y')
    
    version = pecos.__version__
    
    return template.render(**locals())

def _html_template_dashboard(column_names, row_names, content, title, footnote, logo, im_width, datatables, encode):
    
    # if encode == True, encode the images
    img_dic = {}
    if encode:
        for column in column_names:
            for row in row_names:
                try: # not all row/columns have graphics
                    for im in content[row, column]['graphics']:
                       img_encode = base64.b64encode(open(im, "rb").read()).decode("utf-8")
                       img_dic[im] = img_encode 
                except:
                    pass
    
    template = env.get_template('dashboard.html')

    date = datetime.datetime.now()
    datestr = date.strftime('%m/%d/%Y')

    version = pecos.__version__

    return template.render(**locals())

def device_to_client(config):
    """
    Read channels on modbus device, scale and calibrate the values, and store the data in a MySQL database.
    The inputs are provided by a configuration dictionary that describe general information for
    data acquisition and the devices.
    
    Parameters
    ----------
    config : dictionary
        Configuration options, see :ref:`devicetoclient_config`
    """ 
    
    # Extract Database Information
    sec0 = float(datetime.datetime.now().strftime('%s'))
    while True: 
        logging.info('Device to client: '+str(datetime.datetime.now()))
        sec1 = float(datetime.datetime.now().strftime('%s'))
        if sec1 - sec0 >= config['Client']['Interval']:
            run = True
            sec0 = sec1
        else:
            run = False
        
        if run:
            dt = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            labels,dall = [],[]
            retry = config['Client']['Retries']
            for device in config['Devices']:
                # Read channels on modbus device 
                instr = minimalmodbus.Instrument(device['USB'],device['Address'])
                instr.serial.baudrate = device['Baud']
                instr.serial.bytesize = device['Bytes']
                instr.serial.stopbits = device['Stopbits']
                instr.serial.parity = device['Parity']
                
                ds,ls = [],[]
                for data in device['Data']:
                    i = 0
                    while i < retry:
                        l = data['Name']
                        try:
                            d = instr.read_register(data['Channel'], 
                            						numberOfDecimals=data['Scale'], 
                            						functioncode=data['Fcode'], 
                            						signed=data['Signed']) * data['Conversion']
                            break
                        except:
                            if i == retry-1:
                                d = np.nan
                            else:
                                pass
                        i += 1

                    ds.append(d)
                    ls.append(l) 
                
                dall.extend(ds)
                labels.extend(ls)  

            # Add datetime to collected channel values and labels
            dall.extend([dt])
            labels.extend(['datetime'])
            logging.info(ds)
  
            # Convert collected data into pandas DataFrame format
            df = pd.DataFrame(dall).T
            df.columns = labels
            df = df.where((pd.notnull(df)),None)

            # Insert data into database 
            try:
                # Connect to database
                engine = create_engine('mysql://'+config['Client']['Username']+ \
                                       ':'+config['Client']['Password']+'@'+ \
                                        config['Client']['IP']+'/'+ \
                                        config['Client']['Database'])	
                # Write DataFrame to database
                df.to_sql(name=config['Client']['Table'],con=engine, 
                          if_exists='append', index=False) #,dtype = data_type)		
            except:
                pass



"""
The pv module contains custom methods for PV applications.
"""
import pandas as pd
from pecos.metrics import time_integral
import logging

logger = logging.getLogger(__name__)

def insolation(G, tfilter=None, per_day=True):
    """
    Compute insolation defined as:
    
    :math:`H=\int{Gdt}`
    
    where 
    :math:`G` is irradiance and 
    :math:`dt` is the time step between observations.
    The time integral is computed using the trapezoidal rule.
    Results are given in [irradiance units]*seconds.
    
     Parameters
    -----------
    G : pandas DataFrame
        Irradiance time series
        
    tfilter : pandas Series (optional)
        Time filter containing boolean values for each time index
        
    per_day : boolean (optional)
        Flag indicating if the results should be computed per day, default = True
    
    Returns
    -------
    Insolation in a pandas DataFrame
    """
    if type(G) is pd.core.series.Series:
        G = G.to_frame('Irradiance')
        
    H = time_integral(G,  tfilter=tfilter, per_day=per_day)
    
    return H
    
def energy(P, tfilter=None, per_day=True):
    """
    Convert energy defined as:
    
    :math:`E=\int{Pdt}`
    
    where 
    :math:`P` is power and 
    :math:`dt` is the time step between observations.
    The time integral is computed using the trapezoidal rule.
    Results are given in [power units]*seconds.
    
    Parameters
    -----------
    P : pandas DataFrame
        Power time series
         
    tfilter : pandas Series (optional)
        Time filter containing boolean values for each time index
        
    per_day : boolean (optional)
        Flag indicating if the results should be computed per day, default = True
    
    Returns
    -------
    Energy in a pandas DataFrame 
    """
    if type(P) is pd.core.series.Series:
        P = P.to_frame('Power')
        
    E = time_integral(P, tfilter=tfilter, per_day=per_day)
    
    return E

def performance_ratio(E, H_poa, P_ref, G_ref=1000):
    """
    Compute performance ratio defined as:

    :math:`PR=\dfrac{Y_{f}}{Yr} = \dfrac{\dfrac{E}{P_{ref}}}{\dfrac{H_{poa}}{G_{ref}}}`
    
    where 
    :math:`Y_f` is the observed energy (AC or DC) produced by the PV system (kWh) 
    divided by the DC power rating at STC conditions.
    :math:`Y_r` is the plane-of-array insolation (kWh/m2) divided 
    by the reference irradiance (1000 W/m2).
    
    Parameters
    -----------
    E : pandas DataFrame with a single column or pandas Series
        Energy (AC or DC) 
        
    H_poa : pandas DataFrame with a single column or pandas Series
         Plane of array insolation
         
    P_ref : float
        DC power rating at STC conditions
        
    G_ref : float (optional)
        Reference irradiance, default = 1000
        
    Returns
    -------
    Performance ratio in a pandas DataFrame
    """
    logger.info("Compute Performance Ratio")
    
    if type(E) is pd.core.frame.DataFrame:
        E = pd.Series(E.values[:,0], index=E.index)
    if type(H_poa) is pd.core.frame.DataFrame:
        H_poa = pd.Series(H_poa.values[:,0], index=H_poa.index)
        
    Yf = E/P_ref
    Yr = H_poa/G_ref
    PR = Yf/Yr
    
    PR = PR.to_frame('Performance Ratio')
    
    return PR

def normalized_current(I, G_poa, I_sco, G_ref=1000):
    """
    Compute normalized current defined as:

    :math:`NI = \dfrac{\dfrac{I}{I_{sco}}}{\dfrac{G_{poa}}{G_{ref}}}`
    
    where 
    :math:`I` is current, 
    :math:`I_{sco}` is the short circuit current at STC conditions, 
    :math:`G_{poa}` is the plane-of-array irradiance, and 
    :math:`G_{ref}` is the reference irradiance.
    
    Parameters
    -----------
    I : pandas DataFrame with a single column or pandas Series
        Current
        
    G_poa : pandas DataFrame with a single column or pandas Series
         Plane of array irradiance
         
    I_sco : float
        Short circuit current at STC conditions
        
    G_ref : float (optional)
        Reference irradiance, default = 1000
        
    Returns
    -------
    Normalized current in a pandas DataFrame
    """
    logger.info("Compute Normalized Current")
    
    if type(I) is pd.core.frame.DataFrame:
        I = pd.Series(I.values[:,0], index=I.index)
    if type(G_poa) is pd.core.frame.DataFrame:
        G_poa = pd.Series(G_poa.values[:,0], index=G_poa.index)
        
    N = I/I_sco
    D = G_poa/G_ref
    NI = N/D
    
    NI = NI.to_frame('Normalized Current')
    
    return NI
    
def normalized_efficiency(P, G_poa, P_ref, G_ref=1000):
    """
    Compute normalized efficiency defined as:

    :math:`NE = \dfrac{\dfrac{P}{P_{ref}}}{\dfrac{G_{poa}}{G_{ref}}}`
    
    where 
    :math:`P` is the observed power (AC or DC), 
    :math:`P_{ref}` is the DC power rating at STC conditions, 
    :math:`G_{poa}` is the plane-of-array irradiance, and 
    :math:`G_{ref}` is the reference irradiance.
    
    Parameters
    -----------
    P : pandas DataFrame with a single column or pandas Series
        Power (AC or DC) 
        
    G_poa : pandas DataFrame with a single column or pandas Series
         Plane of array irradiance
         
    P_ref : float
        DC power rating at STC conditions
        
    G_ref : float (optional)
        Reference irradiance, default = 1000
        
    Returns
    -------
    Normalized efficiency in a pandas DataFrame
    """
    logger.info("Compute Normalized Efficiency")
    
    if type(P) is pd.core.frame.DataFrame:
        P = pd.Series(P.values[:,0], index=P.index)
    if type(G_poa) is pd.core.frame.DataFrame:
        G_poa = pd.Series(G_poa.values[:,0], index=G_poa.index)
        
    Yf = P/P_ref
    Yr = G_poa/G_ref
    NE = Yf/Yr
    
    NE = NE.to_frame('Normalized Efficiency')
    
    return NE
    
def performance_index(E, E_predicted):
    """
    Compute performance index defined as:
    
    :math:`PI=\dfrac{E}{\hat{E}}`
    
    where 
    :math:`E` is the observed energy from a PV system and  
    :math:`\hat{E}` is the predicted energy over the same time frame.
    :math:`\hat{E}` can be computed using methods in ``pvlib.pvsystem`` 
    and then convert power to energy using ``pecos.pv.enery``.
    
    Unlike with the performance ratio, the performance index should be very 
    close to 1 for a well functioning PV system and should not vary by 
    season due to temperature variations.
    
    Parameters
    -----------
    E : pandas DataFrame with a single column or pandas Series
        Observed energy
    
    E_predicted : pandas DataFrame with a single column or pandas Series
        Predicted energy
        
    Returns
    ---------
    Performance index in a pandas DataFrame
    """
    logger.info("Compute Performance Index")
    
    if type(E) is pd.core.frame.DataFrame:
        E = pd.Series(E.values[:,0], index=E.index)
    if type(E_predicted) is pd.core.frame.DataFrame:
        E_predicted = pd.Series(E_predicted.values[:,0], index=E_predicted.index)
        
    PI = E/E_predicted
    
    PI = PI.to_frame('Performance Index')
    
    return PI

def energy_yield(E, P_ref):
    """
    Compute energy yield is defined as:
    
    :math:`EY=\dfrac{E}{P_{ref}}`
    
    where 
    :math:`E` is the observed energy from a PV system and  
    :math:`P_{ref}` is the DC power rating of the system at STC conditions.
    
    Parameters
    -----------
    E : pandas DataFrame with a single column or pandas Series
        Observed energy
    
    P_ref : float
        DC power rating at STC conditions
        
    Returns
    ---------
    Energy yield in a pandas DataFrame
    """
    logger.info("Compute Energy Yield")
    
    if type(E) is pd.core.frame.DataFrame:
        E = pd.Series(E.values[:,0], index=E.index)

    EY = E/P_ref
    
    EY = EY.to_frame('Energy Yield')
    
    return EY
    
def clearness_index(H_dn, H_ea):
    """
    Compute clearness index defined as:
    
    :math:`Kt=\dfrac{H_{dn}}{H_{ea}}`
    
    where 
    :math:`H_{dn}` is the direct-normal insolation (kWh/m2)
    :math:`H_{ea}` is the extraterrestrial insolation (kWh/m2)
    over the same time frame.
    Extraterrestrial irradiation can be computed using ``pvlib.irradiance.extraradiation``.  
    Irradiation can be converted to insolation using ``pecos.pv.insolation``.
    
    Parameters
    -----------
    H_dn : pandas DataFrame with a single column or pandas Series
        Direct normal insolation
    
    H_ea : pandas DataFrame with a single column or pandas Series
        Extraterrestrial insolation
        
    Returns
    -------
    Clearness index in a pandas DataFrame
    """
    logger.info("Compute Clearness Index")
    
    if type(H_dn) is pd.core.frame.DataFrame:
        H_dn = pd.Series(H_dn.values[:,0], index=H_dn.index)
    if type(H_ea) is pd.core.frame.DataFrame:
        H_ea = pd.Series(H_ea.values[:,0], index=H_ea.index)
        
    Kt = H_dn/H_ea
        
    Kt = Kt.to_frame('Clearness Index')
    
    return Kt



from pecos import monitoring
from pecos import metrics
from pecos import io
from pecos import graphics
from pecos import logger
from pecos import utils
from pecos import pv

__version__ = '0.1.7'

__copyright__ = """Copyright 2016-2017 National Technology & Engineering 
    Solutions of Sandia, LLC (NTESS). Under the terms of Contract 
	DE-NA0003525 with NTESS, the U.S. Government retains certain rights 
	in this software."""

__license__ = "Revised BSD License"



"""
The monitoring module contains the PerformanceMonitoring class used to run
quality control tests and store results.
"""
import pandas as pd
import numpy as np
import re
import datetime
import logging

none_list = ['','none','None','NONE', None, [], {}]

logger = logging.getLogger(__name__)

class PerformanceMonitoring(object):

    def __init__(self):
        """
        PerformanceMonitoring class
        """
        self.df = pd.DataFrame()
        self.trans = {}
        self.tfilter = pd.Series()
        self.test_results = pd.DataFrame(columns=['Variable Name',
                                                'Start Time', 'End Time',
                                                'Timesteps', 'Error Flag'])

    def _setup_data(self, key, rolling_mean):
        """
        Setup DataFrame, by (optionally) extracting a column and/or smoothing
        data using rolling window mean.
        """
        if self.df.empty:
            logger.info("Empty database")
            return

        # Isolate subset if key is not None
        if key is not None:
            try:
                df = self.df[self.trans[key]]
            except:
                logger.warning("Undefined key: " + key)
                return
        else:
            df = self.df

        # Compute moving average
        if rolling_mean > 0:
            rolling_mean_str = str(rolling_mean) + 's' 
            df = df.rolling(rolling_mean_str).mean()
        
        return df
    
    def _generate_test_results(self, df, bound, specs, min_failures, error_prefix):
        """
        Compare DataFrame to bounds to generate a True/False mask where
        True = passed, False = failed.  Append results to test_results.
        """
        
        # Evaluate strings in bound values
        for i in range(len(bound)):
            if bound[i] in none_list:
                bound[i] = None
            elif type(bound[i]) is str:
                bound[i] = self.evaluate_string('', bound[i], specs)
        
        # Lower Bound
        if bound[0] is not None:
            mask = (df < bound[0])
            error_msg = error_prefix+' < lower bound, '+str(bound[0])
            self._append_test_results(mask, error_msg, min_failures) 

        # Upper Bound
        if bound[1] is not None:
            mask = (df > bound[1])
            error_msg = error_prefix+' > upper bound, '+str(bound[1])
            self._append_test_results(mask, error_msg, min_failures) 
 
    def _append_test_results(self, mask, error_msg, min_failures=1, use_mask_only=False): 
        """
        Append QC results to the PerformanceMonitoring object.

        Parameters
        ----------
        mask : pandas DataFrame
            Result from quality control test, boolean values

        error_msg : string
            Error message to store with the QC results

        min_failures : int (optional)
            Minimum number of consecutive failures required for reporting, 
            default = 1

        use_mask_only : boolean  (optional)
            When True, the mask is used directly to determine test 
            results and the variable name is not included in the 
            test_results. When False, the mask is used in combination with 
            pm.df to extract test results. Default = False
        """
        if not self.tfilter.empty:
            mask[~self.tfilter] = False
        if mask.sum(axis=1).sum(axis=0) == 0:
            return
                
        if use_mask_only:
            sub_df = mask 
        else:
            sub_df = self.df[mask.columns]

        # Find blocks
        order = 'col'
        if order == 'col':
            mask = mask.T

        np_mask = mask.values

        start_nans_mask = np.hstack(
            (np.resize(np_mask[:,0],(mask.shape[0],1)),
             np.logical_and(np.logical_not(np_mask[:,:-1]), np_mask[:,1:])))
        stop_nans_mask = np.hstack(
            (np.logical_and(np_mask[:,:-1], np.logical_not(np_mask[:,1:])),
             np.resize(np_mask[:,-1], (mask.shape[0],1))))

        start_row_idx,start_col_idx = np.where(start_nans_mask)
        stop_row_idx,stop_col_idx = np.where(stop_nans_mask)

        if order == 'col':
            temp = start_row_idx; start_row_idx = start_col_idx; start_col_idx = temp
            temp = stop_row_idx; stop_row_idx = stop_col_idx; stop_col_idx = temp
            #mask = mask.T

        block = {'Start Row': list(start_row_idx),
                 'Start Col': list(start_col_idx),
                 'Stop Row': list(stop_row_idx),
                 'Stop Col': list(stop_col_idx)}

        #if sub_df is None:
        #    sub_df = self.df

        for i in range(len(block['Start Col'])):
            length = block['Stop Row'][i] - block['Start Row'][i] + 1
            if length >= min_failures:
                if use_mask_only:
                    var_name = ''
                else:
                    var_name = sub_df.iloc[:,block['Start Col'][i]].name #sub_df.icol(block['Start Col'][i]).name
                    
                frame = pd.DataFrame([var_name,
                    sub_df.index[block['Start Row'][i]],
                    sub_df.index[block['Stop Row'][i]],
                    length, error_msg],
                    index=['Variable Name', 'Start Time', 
                    'End Time', 'Timesteps', 'Error Flag'])
                frame_t = frame.transpose()
                self.test_results = self.test_results.append(frame_t, ignore_index=True)
    
    def add_dataframe(self, df):
        """
        Add DataFrame to the PerformanceMonitoring object.

        Parameters
        -----------
        df : pandas DataFrame
            DataFrame to add to the PerformanceMonitoring object
        """
        temp = df.copy()

        if self.df is not None:
            self.df = temp.combine_first(self.df)
        else:
            self.df = temp.copy()

        # Add identity 1:1 translation dictionary
        trans = {}
        for col in df.columns:
            trans[col] = [col]

        self.add_translation_dictionary(trans)

    def add_translation_dictionary(self, trans):
        """
        Add translation dictionary to the PerformanceMonitoring object.

        Parameters
        -----------
        trans : dictionary
            Translation dictionary
        """
        for key, values in trans.items():
            self.trans[key] = []
            for value in values:
                self.trans[key].append(value)

    def add_time_filter(self, time_filter):
        """
        Add a time filter to the PerformanceMonitoring object.

        Parameters
        ----------
        time_filter : pandas DataFrame with a single column or pandas Series
            Time filter containing boolean values for each time index
        """
        if isinstance(time_filter, pd.DataFrame):
            self.tfilter = pd.Series(data = time_filter.values[:,0], index = self.df.index)
        else:
            self.tfilter = time_filter

#    def add_signal(self, col_name, data):
#        """
#        Add signal to the PerformanceMonitoring DataFrame.
#
#        Parameters
#        -----------
#        col_name : string
#            Column name to add to translation dictionary
#
#        data : pandas DataFrame or pandas Series
#            Data to add to df
#        """
#        if type(data) is pd.core.series.Series:
#            data = data.to_frame(col_name)
#        if type(data) is not pd.core.frame.DataFrame:
#            logger.warning("Add signal failed")
#            return
#
#        if col_name in self.trans.keys():
#            logger.info(col_name + ' already exists in trans')
#            return
#        for col in data.columns.values.tolist():
#            if col in self.df.columns.values.tolist():
#                logger.info(col + ' already exists in df')
#                return
#        try:
#            self.trans[col_name] = data.columns.values.tolist()
#            #self.df[df.columns] = df
#            for col in data.columns:
#                self.df[col] = data[col]
#        except:
#            logger.warning("Add signal failed: " + col_name)
#            return

    def check_timestamp(self, frequency, expected_start_time=None,
                        expected_end_time=None, min_failures=1,
                        exact_times=True):
        """
        Check time series for missing, non-monotonic and duplicate
        timestamps.

        Parameters
        ----------
        frequency : int
            Expected time series frequency, in seconds

        expected_start_time : Timestamp (optional)
            Expected start time. If not specified, the minimum timestamp
            is used

        expected_end_time : Timestamp (optional)
            Expected end time. If not specified, the maximum timestamp
            is used

        min_failures : int (optional)
            Minimum number of consecutive failures required for
            reporting, default = 1

        exact_times : bool (optional)
            Controls how missing times are checked. 
            If True, times are expected to occur at regular intervals 
            (specified in frequency) and the DataFrame is reindexed to match 
            the expected frequency.
            If False, times only need to occur once or more within each 
            interval (specified in frequency) and the DataFrame is not 
            reindexed.
        """
        logger.info("Check timestamp")

        if self.df.empty:
            logger.info("Empty database")
            return
        if expected_start_time is None:
            expected_start_time = min(self.df.index)
        if expected_end_time is None:
            expected_end_time = max(self.df.index)

        rng = pd.date_range(start=expected_start_time, end=expected_end_time,
                            freq=str(int(frequency*1e6)) + 'us') # microseconds

        # Check to see if timestamp is monotonic
#        mask = pd.TimeSeries(self.df.index).diff() < 0
        mask = pd.Series(self.df.index).diff() < pd.Timedelta('0 days 00:00:00')
        mask.index = self.df.index
        mask[mask.index[0]] = False
        mask = pd.DataFrame(mask)
        mask.columns = [0]

        self._append_test_results(mask, 'Nonmonotonic timestamp',
                                 use_mask_only=True,
                                 min_failures=min_failures)

        # If not monotonic, sort df by timestamp
        if not self.df.index.is_monotonic:
            self.df = self.df.sort_index()

        # Check for duplicate timestamps
#        mask = pd.TimeSeries(self.df.index).diff() == 0
        mask = pd.Series(self.df.index).diff() == pd.Timedelta('0 days 00:00:00')
        mask.index = self.df.index
        mask[mask.index[0]] = False
        mask = pd.DataFrame(mask)
        mask.columns = [0]
        mask['TEMP'] = mask.index # remove duplicates in the mask
        mask.drop_duplicates(subset='TEMP', keep='last', inplace=True)
        del mask['TEMP']

        # Drop duplicate timestamps (this has to be done before the
        # results are appended)
        self.df['TEMP'] = self.df.index
        #self.df.drop_duplicates(subset='TEMP', take_last=False, inplace=True)
        self.df.drop_duplicates(subset='TEMP', keep='first', inplace=True)

        self._append_test_results(mask, 'Duplicate timestamp',
                                 use_mask_only=True,
                                 min_failures=min_failures)
        del self.df['TEMP']
        
        if exact_times:
            temp = pd.Index(rng)
            missing = temp.difference(self.df.index).tolist()
            # reindex DataFrame
            self.df = self.df.reindex(index=rng)
            mask = pd.DataFrame(data=self.df.shape[0]*[False],
                                index=self.df.index)
            mask.loc[missing] = True
            self._append_test_results(mask, 'Missing timestamp',
                                 use_mask_only=True,
                                 min_failures=min_failures)
        else:
            # uses pandas >= 0.18 resample syntax
            df_index = pd.DataFrame(index=self.df.index)
            df_index[0]=1 # populate with placeholder values
            mask = df_index.resample('{}s'.format(frequency)).count() == 0
            self._append_test_results(mask, 'Missing timestamp',
                                 use_mask_only=True,
                                 min_failures=min_failures)
        
    def check_range(self, bound, key=None, specs={}, rolling_mean=0, min_failures=1):
        """
        Check bounds on data.

        Parameters
        ----------
        bound : list of floats
            [lower bound, upper bound], None can be used in place of a lower 
            or upper bound

        key : string (optional)
            Translation dictionary key.  If not specified, all columns are 
            used in the test.

        specs : dictionary (optional)
            Constants used in bound

        rolling_mean : int (optional)
            Size of the moving window (in seconds) used to smooth data using a
            rolling mean before the test is run, default = 0 (i.e., not used)

        min_failures : int (optional)
            Minimum number of consecutive failures required for reporting, 
            default = 1
        """
        logger.info("Check data range")

        df = self._setup_data(key, rolling_mean)
        if df is None:
            return
        
        error_prefix = 'Data'
            
        self._generate_test_results(df, bound, specs, min_failures, error_prefix)
    
    def check_increment(self, bound, key=None, specs={}, increment=1, 
                        absolute_value=True, rolling_mean=0, min_failures=1):
        """
        Check bounds on the difference between data values separated by a set 
        increment.

        Parameters
        ----------
        bound : list of floats
            [lower bound, upper bound], None can be used in place of a lower 
            or upper bound

        key : string (optional)
            Translation dictionary key. If not specified, all columns are 
            used in the test.

        specs : dictionary (optional)
            Constants used in bound

        increment : int (optional)
            Time step shift used to compute difference, default = 1

        absolute_value : boolean (optional)
            Take the absolute value of the increment data, default = True

        rolling_mean : int (optional)
            Size of the moving window (in seconds) used to smooth data using a
            rolling mean before the test is run, default = 0 (i.e., not used)

        min_failures : int (optional)
            Minimum number of consecutive failures required for reporting,
            default = 1
        """
        logger.info("Check increment range")

        df = self._setup_data(key, rolling_mean)
        if df is None:
            return
        
        if df.isnull().all().all():
            logger.warning("Check increment range failed (all data is Null): " + key)
            return
        
        # Compute interval
        if absolute_value:
            df = np.abs(df.diff(periods=increment))
        else:
            df = df.diff(periods=increment)
        
        if absolute_value:
            error_prefix = '|Increment|'
        else:
            error_prefix = 'Increment'
            
        self._generate_test_results(df, bound, specs, min_failures, error_prefix)
      
    def check_delta(self, bound, key=None, specs={}, window=3600, 
                    absolute_value=True, rolling_mean=0, min_failures=1):
        """
        Check bounds on the difference between max and min data values within 
		  a rolling window (Note, this method is currently NOT efficient for large 
        data sets (> 100000 pts) because it uses df.rolling().apply() to find 
        the position of the min and max). This method requires pandas 0.23 or greater.

        Parameters
        ----------
        bound : list of floats
            [lower bound, upper bound], None can be used in place of a lower 
            or upper bound

        key : string (optional)
            Translation dictionary key. If not specified, all columns are used 
            in the test.

        specs : dictionary (optional)
            Constants used in bound

        window : int (optional)
            Size of the moving window (in seconds) used to compute delta, 
            default = 3600

        absolute_value : boolean (optional)
            Take the absolute value of delta, default = True

        rolling_mean : int (optional)
            Size of the moving window (in seconds) used to smooth data using a
            rolling mean before the test is run, default = 0 (i.e., not used)

        min_failures : int (optional)
            Minimum number of consecutive failures required for reporting, 
            default = 1
        """
        logger.info("Check delta (max-min) range")

        df = self._setup_data(key, rolling_mean)
        if df is None:
            return
        
        window_str = str(int(window*1e6)) + 'us'

        def f(data=None, method=None):
            if data.notnull().sum() < 2: # there has to be at least two numbers
                return np.nan
            else:
                if method == 'idxmin':
                    # Can't return a timestamp, convert to num, then back to timestamp
                    return data.idxmin().value 
                elif method == 'min':
                    return data.min()
                elif method == 'idxmax':
                    return data.idxmax().value
                elif method == 'max':
                    return data.max()
                else:
                    return np.nan
                
        tmin_df = df.rolling(window_str).apply(lambda x: f(x, 'idxmin'), raw=False) # raw = False passes a Series
        tmin_df = tmin_df.astype('datetime64[ns]')
        # Note, the next line should be replaced with df.rolling(window_str).min(), 
        # but the solution is not the same with pandas 0.23
        min_df = df.rolling(window_str).apply(lambda x: f(x, 'min'), raw=False)
        
        tmax_df = df.rolling(window_str).apply(lambda x: f(x, 'idxmax'), raw=False)
        tmax_df = tmax_df.astype('datetime64[ns]')
        # Same note as above, for max
        max_df = df.rolling(window_str).apply(lambda x: f(x, 'max'), raw=False)
        
        diff_df = max_df - min_df
        if not absolute_value:
            reverse_order = tmax_df < tmin_df 
            diff_df[reverse_order] = -diff_df[reverse_order]

        if absolute_value:
            error_prefix = '|Delta|'
        else:
            error_prefix = 'Delta'
        
        # Evaluate strings for bound values
        for i in range(len(bound)):
            if bound[i] in none_list:
                bound[i] = None
            elif type(bound[i]) is str:
                bound[i] = self.evaluate_string('', bound[i], specs)
        
        def extract_exact_position(mask1, tmin_df, tmax_df):
            mask2 = pd.DataFrame(False, columns=mask1.columns, index=mask1.index)
            # Loop over t, col in mask1 where condition is True
            for t,col in list(mask1[mask1 > 0].stack().index): 
                # set the initially flaged location to False
                mask2.loc[t,col] = False 
                # extract the start and end time
                start_time = tmin_df.loc[t,col]
                end_time = tmax_df.loc[t,col]
                # update mask2
                if start_time < end_time:
                    mask2.loc[start_time:end_time,col] = True # set the time between max and min to true
                else:
                    mask2.loc[end_time:start_time,col] = True # set the time between max and min to true
            return mask2
        
        # Lower Bound
        if bound[0] is not None:
            mask = (diff_df < bound[0])
            if not self.tfilter.empty:
                mask[~self.tfilter] = False
            if mask.sum(axis=1).sum(axis=0) > 0:
                mask = extract_exact_position(mask, tmin_df, tmax_df)
                self._append_test_results(mask, error_prefix+' < lower bound, '+str(bound[0]), 
                                         min_failures=min_failures) 

        # Upper Bound
        if bound[1] is not None:
            mask = (diff_df > bound[1])
            if not self.tfilter.empty:
                mask[~self.tfilter] = False
            if mask.sum(axis=1).sum(axis=0) > 0:
                mask = extract_exact_position(mask, tmin_df, tmax_df)
                self._append_test_results(mask, error_prefix+' > upper bound, '+str(bound[1]), 
                                         min_failures=min_failures) 
                
    def check_outlier(self, bound, key=None, specs={}, window=3600, 
                        absolute_value=True, rolling_mean=0, min_failures=1):
        """
        Check bounds on normalized data within a moving window to find outliers.
        The bound is specified in standard deviations.
        Data normalized using (data-mean)/std.

        Parameters
        ----------
        bound : list of floats
            [lower bound, upper bound], None can be used in place of a lower 
            or upper bound

        key : string (optional)
            Translation dictionary key. If not specified, all columns are used 
            in the test.

        specs : dictionary (optional)
            Constants used in bound

        window : int or None (optional)
            Size of the moving window (in seconds) used to normalize data, 
            default = 3600.  If window is set to None, data is normalized using 
            the entire data sets mean and standard deviation (column by column).

        absolute_value : boolean (optional)
            Take the absolute value the normalized data, default = True

        rolling_mean : int (optional)
            Size of the moving window (in seconds) used to smooth data using a
            rolling mean before the test is run, default = 0 (i.e., not used)

        min_failures : int (optional)
            Minimum number of consecutive failures required for reporting, 
            default = 1
        """
        logger.info("Check for outliers")

        df = self._setup_data(key, rolling_mean)
        if df is None:
            return

        # Compute normalized data
        if window is not None:
            window_str = str(int(window*1e6)) + 'us'
            df = (df - df.rolling(window_str).mean())/df.rolling(window_str).std()
        else:
            df = (df - df.mean())/df.std()
        if absolute_value:
            df = np.abs(df)
        df.replace([np.inf, -np.inf], np.nan, inplace=True)
        
        if absolute_value:
            error_prefix = '|Outlier|'
        else:
            error_prefix = 'Outlier'
        
        #df[df.index[0]:df.index[0]+datetime.timedelta(seconds=window)] = np.nan
             
        self._generate_test_results(df, bound, specs, min_failures, error_prefix)
      
    def check_missing(self, key=None, min_failures=1):
        """
        Check for missing data

        Parameters
        ----------
        key : string (optional)
            Translation dictionary key. If not specified, all columns are used 
            in the test.

        min_failures : int (optional)
            Minimum number of consecutive failures required for reporting, 
            default = 1
        """
        logger.info("Check for missing data")

        df = self._setup_data(key, 0)
        if df is None:
            return
        
        # Extract missing data
        mask = pd.isnull(df) # checks for np.nan, np.inf

        missing_timestamps = self.test_results[
                self.test_results['Error Flag'] == 'Missing timestamp']
        for index, row in missing_timestamps.iterrows():
            mask.loc[row['Start Time']:row['End Time']] = False

        self._append_test_results(mask, 'Missing data', min_failures=min_failures)

    def check_corrupt(self, corrupt_values, key=None, min_failures=1):
        """
        Check for corrupt data.

        Parameters
        ----------
        corrupt_values : list of floats
            List of corrupt data values

        key : string (optional)
            Translation dictionary key. If not specified, all columns are used 
            in the test.

        min_failures : int (optional)
            Minimum number of consecutive failures required for reporting, 
            default = 1
        """
        logger.info("Check for corrupt data")

        df = self._setup_data(key, 0)
        if df is None:
            return
        
        # Extract corrupt data
        mask = pd.DataFrame(data = np.zeros(df.shape), index = df.index, columns = df.columns, dtype = bool) # all False
        for i in corrupt_values:
            mask = mask | (df == i)
        self.df[mask] = np.nan
               
        self._append_test_results(mask, 'Corrupt data', min_failures=min_failures)

    def evaluate_string(self, col_name, string_to_eval, specs={}):
        """
        Returns the evaluated Python equation written as a string (BETA).
        For each {keyword} in string_to_eval,
        {keyword} is first expanded to self.df[self.trans[keyword]],
        if that fails, then {keyword} is expanded to specs[keyword].

        Parameters
        ----------
        col_name : string
            Column name for the new signal

        string_to_eval : string
            String to evaluate

        specs : dictionary (optional)
            Constants used as keywords

        Returns
        --------
        pandas DataFrame or pandas Series with the evaluated string
        """

        match = re.findall(r"\{(.*?)\}", string_to_eval)
        for m in set(match):
            m = m.replace('[','') # check for list

            if m == 'ELAPSED_TIME':
                ELAPSED_TIME = self.get_elapsed_time()
                string_to_eval = string_to_eval.replace("{"+m+"}",m)
            elif m == 'CLOCK_TIME':
                CLOCK_TIME = self.get_clock_time()
                string_to_eval = string_to_eval.replace("{"+m+"}",m)
            else:
                try:
                    self.df[self.trans[m]]
                    datastr = "self.df[self.trans['" + m + "']]"
                    string_to_eval = string_to_eval.replace("{"+m+"}",datastr)
                except:
                    try:
                        specs[m]
                        datastr = "specs['" + m + "']"
                        string_to_eval = string_to_eval.replace("{"+m+"}",datastr)
                    except:
                        pass

        try:
            signal = eval(string_to_eval)
            if type(signal) is tuple: # A tuple of series
                col_name = [col_name + " " + str(i+1)  for i in range(len(signal))]
                signal = pd.concat(signal, axis=1)
                signal.columns = col_name
                signal.index = self.df.index
            elif type(signal) is float:
                signal = signal
            else:
                signal = pd.DataFrame(signal)
                if len(signal.columns) == 1:
                    signal.columns = [col_name]
                else:
                    signal.columns = [col_name + " " + str(i+1)  for i in range(signal.shape[1])]
                signal.index = self.df.index
        except:
            signal = None
            logger.warning("Insufficient data for Composite Signals: " + col_name + ' -- ' + string_to_eval)

        return signal

    def get_elapsed_time(self):
        """
        Returns the elapsed time in seconds for each Timestamp in the 
        DataFrame index.

        Returns
        --------
        pandas DataFrame with elapsed time of the DataFrame index
        """
        elapsed_time = ((self.df.index - self.df.index[0]).values)/1000000000 # convert ns to s
        elapsed_time = pd.DataFrame(data=elapsed_time, index=self.df.index, dtype=int)

        return elapsed_time

    def get_clock_time(self):
        """
        Returns the time of day in seconds past midnight for each Timestamp 
        in the DataFrame index.

        Returns
        --------
        pandas DataFrame with clock time of the DataFrame index
        """

        secofday = self.df.index.hour*3600 + \
                   self.df.index.minute*60 + \
                   self.df.index.second + \
                   self.df.index.microsecond/1000000.0
        clock_time = pd.DataFrame(secofday, index=self.df.index)

        return clock_time

    def get_test_results_mask(self, key=None):
        """
        Return a mask of data-times that failed a quality control test.

        Parameters
        -----------
        key : string (optional)
            Translation dictionary key. If not specified, all columns are used

        Returns
        --------
        pandas DataFrame containing boolean values for each data point, True =
        data point pass all tests, False = data point did not pass at least 
        one test (or data is NaN).
        """
        if self.df.empty:
            logger.info("Empty database")
            return

        if key is not None:
            try:
                df = self.df[self.trans[key]]
            except:
                logger.warning("Key not in DataFrame")
                return
        else:
            df = self.df

        test_results_mask = ~pd.isnull(df) # False if NaN
        for i in self.test_results.index:
            variable = self.test_results.loc[i, 'Variable Name']
            start_date = self.test_results.loc[i, 'Start Time']
            end_date = self.test_results.loc[i, 'End Time']
            if variable in test_results_mask.columns:
                try:
                    test_results_mask.loc[start_date:end_date,variable] = False
                except:
                    pass
                
        return test_results_mask



"""
The graphics module contains functions to generate scatter, time series, and 
heatmap plots for reports.
"""
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
try:
    import plotly
except:
    pass
import textwrap
import os
import logging

try:
    from nose.tools import nottest as _nottest
except ImportError:
    def _nottest(afunction):
        return afunction
        
logger = logging.getLogger(__name__)

def plot_scatter(x,y,xaxis_min=None, xaxis_max=None, yaxis_min=None, 
                 yaxis_max=None, title=None, figsize=(7.0, 3.0)):
    """
    Create a scatter plot.  If x and y have the same number of columns, then 
    the columns of x are plotted against the corresponding columns of y, in order.
    If x (or y) has 1 column, then that column of data is plotted against all
    the columns in y (or x).
    
    Parameters
    ----------
    x : pandas DataFrame
        X data
    
    y : pandas DataFrame
        Y data
    
    xaxis_min : float (optional)
        X-axis minimum, default = None (autoscale)       
        
    xaxis_max : float (optional)
        X-axis maximum, default = None (autoscale)
        
    yaxis_min : float (optional)
        Y-axis minimum, default = None (autoscale)     
        
    yaxis_max : float (optional)
        Y-axis maximum, default = None (autoscale)        
    
    title : string (optional)
        Title, default = None
    
    figsize : tuple (optional)
        Figure size, default = (7.0, 3.0)
    """
    
    plt.figure(figsize = figsize)
    ax = plt.gca()

    try:
        if x.shape[1] == y.shape[1]:
            for i in range(x.shape[1]):
                plt.plot(x.iloc[:,i],y.iloc[:,i], '.', markersize=3) 
                plt.xticks(rotation='vertical')
        elif x.shape[1] != y.shape[1]:
            if x.shape[1] == 1:
                for col in y.columns:
                    plt.plot(x,y[col], '.', markersize=3) 
                    plt.xticks(rotation='vertical')
            elif y.shape[1] == 1:
                for col in x.columns:
                    plt.plot(x[col],y, '.', markersize=3)
                    plt.xticks(rotation='vertical')
    except:
        plt.text(0.3,0.5,'Insufficient Data', fontsize=8)

    # Format axis
    xmin_plt, xmax_plt = plt.xlim()
    ymin_plt, ymax_plt = plt.ylim()
    if xaxis_min is None:
        xaxis_min = xmin_plt
    if xaxis_max is None:
        xaxis_max = xmax_plt
    if yaxis_min is None:
        yaxis_min = ymin_plt
    if yaxis_max is None:
        yaxis_max = ymax_plt
    plt.xlim((xaxis_min, xaxis_max))
    plt.ylim((yaxis_min, yaxis_max))
    if title:
        plt.title(title)
    ax.tick_params(axis='both', labelsize=8)
    box = ax.get_position()
    ax.set_position([box.x0, box.y0+0.15, box.width, box.height*0.75])

def plot_timeseries(data, tfilter=None, test_results_group=None, xaxis_min=None, 
                    xaxis_max=None, yaxis_min=None, yaxis_max=None, title=None,
                    figsize=(7.0, 3.0)):
    """
    Create a time series plot using each column in the DataFrame.
    
    Parameters
    ----------
    data : pandas DataFrame
        Data, indexed by time
        
    tfilter : pandas Series (optional)
        Boolean values used to include time filter in the plot, default = None 
        
    test_results_group : pandas DataFrame (optional)
        Test results for a particular variable.
        default = None 
    
    xaxis_min : float (optional)
        X-axis minimum, default = None (autoscale)        
        
    xaxis_max : float (optional)
        X-axis maximum, default = None (autoscale)    
        
    yaxis_min : float (optional)
        Y-axis minimum, default = None (autoscale)            
        
    yaxis_max : float (optional)
        Y-axis maximum, default = None (autoscale)
        
    title : string (optional)
        Title, default = None
    
    figsize : tuple (optional)
        Figure size, default = (7.0, 3.0)
    """
    
    plt.figure(figsize = figsize)
    ax = plt.gca()
    
    try:
        # plot time series
        if isinstance(data, pd.Series):
            data.plot(ax=ax, linewidth=1, grid=False, legend=False, color='k', 
                      fontsize=8, rot=90, label='Data', x_compat=True)
        else:
            data.plot(ax=ax, linewidth=1, grid=False, legend=False, 
                      fontsize=8, rot=90, label='Data')
    
        if tfilter is not None:
            # add tfilter        
            temp = np.where(tfilter - tfilter.shift())
            temp = np.append(temp[0],len(tfilter)-1)
            count = 0
            for i in range(len(temp)-1):
                if tfilter[temp[i]] == 0:
                    if count == 0:
                        ax.axvspan(data.index[temp[i]], data.index[temp[i+1]], 
                                   facecolor='k', alpha=0.2, label='Time filter')
                        count = count+1
                    else:
                        ax.axvspan(data.index[temp[i]], data.index[temp[i+1]], 
                                   facecolor='k', alpha=0.2)     
        
        # add errors 
        try:
            if test_results_group.empty:
                test_results_group = None
        except:
            pass
        if test_results_group is not None:
            key2 = test_results_group['Error Flag']
            grouped2 = test_results_group.groupby(key2)
            
            for error_flag in key2.unique():
                test_results_group2 = grouped2.get_group(error_flag)
                
                error_label = '\n'.join(textwrap.wrap(error_flag, 30))
                if len(test_results_group2.index.values) > 4:
                    warning_label = '\n'.join(textwrap.wrap('Warning ' + 
                        str(test_results_group2.index.values[0:4]).strip('[]') + '...', 30))
                else:
                    warning_label = '\n'.join(textwrap.wrap('Warning ' + 
                        str(test_results_group2.index.values).strip('[]'), 30))
                error_label = error_label + '\n' + warning_label
                
                date_idx2 = np.array([False]*len(data.index))
                for row2 in range(len(test_results_group2.index)):
                    s_index = test_results_group2.columns.get_loc("Start Time")
                    e_index = test_results_group2.columns.get_loc("End Time")
                    date_idx2 = date_idx2 + ((data.index >= test_results_group2.iloc[row2,s_index]) & 
                                             (data.index <= test_results_group2.iloc[row2,e_index]))
                
                if sum(date_idx2) == 0:
                    continue
                
                data2 = data[date_idx2]
                if error_flag in ['Duplicate timestamp', 'Missing data', 
                                  'Corrupt data', 'Nonmonotonic timestamp']:
                    continue
                if "Data" in error_flag:
                    color='r'
                elif "Delta" in error_flag:
                    color = 'g'
                else: # Outlier
                    color = 'b'
                try:
                    ax.scatter(data2.index, data2.values, c=color, marker='+', 
                               linewidths=1, label=error_label)   
                except:
                    ax.scatter(data2.index[0], data2.values[0], c=color, marker='+', 
                               linewidths=1, label=error_label) 
        
        # Format axis
        xmin_plt, xmax_plt = plt.xlim()
        ymin_plt, ymax_plt = plt.ylim()
        if tfilter is not None:
            ymin_plt = np.nanmin(data[tfilter].values)
            ymax_plt = np.nanmax(data[tfilter].values)
        if np.abs(ymin_plt - ymax_plt) < 0.01:
            ymin_plt, ymax_plt = plt.ylim()
    except:
        plt.text(0.3,0.5,'Insufficient Data', fontsize=8)
        xmin_plt, xmax_plt = plt.xlim()
        ymin_plt, ymax_plt = plt.ylim()
    
    # Format axis
    y_range = (ymax_plt - ymin_plt)
    if xaxis_min is None:
        xaxis_min = xmin_plt
    if xaxis_max is None:
        xaxis_max = xmax_plt
    if yaxis_min is None:
        yaxis_min = ymin_plt-y_range/10
    if yaxis_max is None:
        yaxis_max = ymax_plt+y_range/10
    plt.xlim((xaxis_min, xaxis_max))
    plt.ylim((yaxis_min, yaxis_max))
    if title:
        plt.title(title)
    ax.get_yaxis().get_major_formatter().set_useOffset(False)
    ax.tick_params(axis='both', labelsize=8)
    plt.xlabel('Time', fontsize=8)
    box = ax.get_position()
    ax.set_position([box.x0, box.y0+0.15, box.width, box.height*0.75])

def plot_interactive_timeseries(data, xaxis_min=None, xaxis_max=None, yaxis_min=None, 
                 yaxis_max=None, title=None, filename=None, auto_open=True):
    """
    Create a basic interactive time series graphic using plotly.  Many more 
    options are available, see https://plot.ly for more details.
    
    Parameters
    ----------
    data : pandas DataFrame
        Data, indexed by time
    
    xaxis_min : float (optional)
        X-axis minimum, default = None (autoscale)       
        
    xaxis_max : float (optional)
        X-axis maximum, default = None (autoscale)
        
    yaxis_min : float (optional)
        Y-axis minimum, default = None (autoscale)     
        
    yaxis_max : float (optional)
        Y-axis maximum, default = None (autoscale)        
    
    title : string (optional)
        Title, default = None
        
    filename : string, optional
        HTML file name, default = None (file will be named temp-plot.html)
    
    auto_open : boolean (optional)
        Flag indicating if HTML graphic is opened, default = True
    """
    
    layout = dict(hovermode = 'closest')
    layout = dict(title=title, hovermode = 'closest',
                  xaxis=dict(range=[xaxis_min,xaxis_max]),
                  yaxis=dict(range=[yaxis_min,yaxis_max]))
    plotly_data = []
    for col in data.columns:
        trace = plotly.graph_objs.Scatter(x=data.index.tz_localize(None), y=data.loc[:,col], name = col)
        plotly_data.append(trace)
    fig = dict(data=plotly_data, layout=layout)
    if filename:
        plotly.offline.plot(fig, filename=filename, auto_open=auto_open)  
    else:
        plotly.offline.plot(fig, auto_open=auto_open)  
        
def plot_heatmap(data, colors=[(0.75, 0.15, 0.15), (1, 0.75, 0.15), (0.15, 0.75, 0.15)], 
                               nColors=12, cmap=None, vmin=None, vmax=None, 
                               show_axis=False, title=None, figsize=(5.0, 5.0)): 
    """ 
    Create a heatmap.  Default color scheme is red to yellow to green with 12 
    colors.  This function can be used to generate dashboards with simple color 
    indicators in each cell (to remove borders use bbox_inches='tight' and 
    pad_inches=0 when saving the image).
    
    Parameters
    -----------
    data : pandas DataFrame or numpy ndarray
        Data
    
    colors : list (optional)
        List of colors, colors can be specified in any way understandable by 
        matplotlib.colors.ColorConverter.to_rgb().
        Default is red to yellow to green.
    
    num_colors : int (optional)
        Number of colors in the colormap, default = 12
        
    cmap : string (optional)
        Colormap, default = None. Overrides colors and num_colors listed above.
        
    vmin : float (optional)
        Colomap minimum, default = None (autoscale) 
    
    vmax : float (optional)
        Colomap maximum, default = None (autoscale) 
    
    title : string (optional)
        Title, default = None
    
    figsize : tuple (optional)
        Figure size, default = (5.0, 5.0)
    """
    if type(data) is pd.core.frame.DataFrame:
            data = data.values
            
    if not cmap:
        from matplotlib.colors import LinearSegmentedColormap
        cmap = LinearSegmentedColormap.from_list(name='custom', colors=colors, N=nColors)
    
    plt.figure(figsize = figsize)
    fig = plt.imshow(data, cmap=cmap, aspect='equal', vmin=vmin, vmax=vmax)
    if not show_axis:
        plt.axis('off')
        fig.axes.get_xaxis().set_visible(False)
        fig.axes.get_yaxis().set_visible(False)
    if title:
        plt.title(title)
    plt.tight_layout()

def plot_doy_heatmap(data, cmap='nipy_spectral', vmin=None, vmax=None, 
                     overlay=None, title=None, figsize=(7.0, 3.0)):
    """
    Create a day-of-year (X-axis) vs. time-of-day (Y-axis) heatmap.
    
    Parameters
    ----------
    data : pandas DataFrame or pandas Series
        Data (single column), indexed by time
        
    cmap : string (optional)
        Colomap, default = nipy_spectral
        
    vmin : float (optional)
        Colomap minimum, default = None (autoscale)
    
    vmax : float (optional)
        Colomap maximum, default = None (autoscale)
        
    overlay : pandas DataFrame (optional)
        Data to overlay on the heatmap.  
        Time index should be in day-of-year (X-axis) 
        Values should be in time-of-day in minutes (Y-axis)
    
    title : string (optional)
        Title, default = None
        
    figsize : tuple (optional)
        Figure size, default = (7.0, 3.0)
    """
    
    if type(data) is pd.core.series.Series:
        data = data.to_frame()
            
    # Convert data to a pivot table
    col_name = data.columns[0]
    data['X'] = data.index.dayofyear
    data['Y'] = data.index.hour*60 + \
                data.index.minute + \
                data.index.second/60 + \
                data.index.microsecond/(60*1000000.0)
    piv = pd.pivot_table(data,values=col_name,index=['Y'],columns=['X'],fill_value=np.NaN)
    
    # Create the heatmap
    plt.figure(figsize = figsize)
    fig, ax = plt.subplots(figsize=figsize)
    im = ax.imshow(piv, cmap=cmap, aspect='auto', vmin=vmin, vmax=vmax,
                   extent=[data['X'].min()-0.5,data['X'].max()+0.5, 
                           data['Y'].max()-0.5,data['Y'].min()+0.5])
    fig.colorbar(im, ax=ax)
    
    # Add overlay
    if type(overlay) is pd.core.frame.DataFrame:
        overlay.plot(ax=ax)
        
    # Add title and labels
    if title:
        ax.set_title(title)
    ax.set_xlabel("Day of the year")
    ax.set_ylabel("Time of day (minutes)")
    plt.tight_layout()
    
@_nottest
def plot_test_results(filename_root, pm, image_format='png', dpi=500, 
                      figsize=(7.0, 3.0)):
    """
    Create test results graphics which highlight data points that
    failed a quality control test.

    Parameters
    ----------
    filename_root : string
        Filename root, with full path.  
        Each grpahics filename is appended with an integer.
        For example, filename_root = 'C:\\\\pecos\\\\results\\\\test' will 
        generate a file named 'C:\\\\pecos\\\\results\\\\test0.png'.
        The directory ''C:\\\\pecos\\\\results' must exist.

    pm : PerformanceMonitoring object
        Contains data (pm.df) and test results (pm.test_results)
        
    image_format : string  (optional)
        Image format, default = 'png'
    
    dpi : int (optional)
        DPI resolution, default = 500
        
    figsize : tuple (optional)
        Figure size, default = (7.0, 3.0)
    """
    
    filename_root = os.path.abspath(filename_root)
    
    # Colect file names
    test_results_graphics = []
    
    if pm.test_results.empty:
        return test_results_graphics

    graphic = 0
    
    pm.test_results.sort_values(list(pm.test_results.columns), inplace=True)
    pm.test_results.index = np.arange(1, pm.test_results.shape[0]+1)
    
    # Remove specific error flags
    remove_error_flags = ['Duplicate timestamp', 
                          'Missing data', 
                          'Corrupt data', 
                          'Missing timestamp', 
                          'Nonmonotonic timestamp']
    test_results = pm.test_results[-pm.test_results['Error Flag'].isin(remove_error_flags)]
    grouped = test_results.groupby(['Variable Name'])

    for col_name, test_results_group in grouped:
        logger.info("Creating graphic for " + col_name)
        
        plot_timeseries(pm.df[col_name], pm.tfilter, 
                        test_results_group=test_results_group, figsize = figsize)

        ax = plt.gca()
        box = ax.get_position()
        ax.set_position([box.x0, box.y0, box.width*0.65, box.height])
        plt.legend(loc='center left', bbox_to_anchor=(1, 0.5), fontsize=8)
        plt.title(col_name, fontsize=8)
        
        filename = filename_root + str(graphic) + '.' + image_format
        test_results_graphics.append(filename)
        plt.savefig(filename, format=image_format, dpi=dpi)
            
        graphic = graphic + 1
        plt.close()

    return test_results_graphics



"""
The utils module contains helper functions.
"""
import pandas as pd
import numpy as np
from pandas.tseries.frequencies import to_offset
import logging
import os

logger = logging.getLogger(__name__)
        
def round_index(dt, frequency, how='nearest'):
    """
    Round datetime index.
    
    Parameters
    ----------
    dt : DatetimeIndex
        Time series index
    
    frequency : int
        Expected time series frequency, in seconds
    
    how : string (optional)
        Method for rounding, default = 'nearest'.  Options include:
        
        - nearest = round the index to the nearest expected integer
        - floor= round the index to the largest expected integer such that the integer <= index
        - ceiling = round the index to the smallest expected integer such that the integer >= index
        
    Returns
    -------
    Rounded time series index, DatetimeIndex
    """
    freq=str(frequency) + 's'
    freq = to_offset(freq).nanos
    
    if how=='nearest':
        rounded_dt = pd.DatetimeIndex(((np.round(dt.asi8/(float(freq)))*freq).astype(np.int64)))
    elif how=='floor':
        rounded_dt = pd.DatetimeIndex(((np.floor(dt.asi8/(float(freq)))*freq).astype(np.int64)))
    elif how=='ceiling':
        rounded_dt = pd.DatetimeIndex(((np.ceil(dt.asi8/(float(freq)))*freq).astype(np.int64)))
    else:
        logger.info("Invalid input, index not rounded")
        rounded_dt = dt

    return rounded_dt

def convert_html_to_image(html_filename, image_filename, image_format='png', quality=100, zoom=1):
    """
    Convert html file to image file using wkhtmltoimage.
    See http://wkhtmltopdf.org/ for more information.
    
    Parameters
    ----------
    html_filename : string
        HTML file name, with full path
    
    image_filename : string
        Image file name, with full path
    
    image_format : string  (optional)
        Image format, default = 'png'
    
    quality : int (optional)
        Image quality, default = 100
    
    zoom : int (optional)
        Zoom factor, default = 1
    """
    os.system('wkhtmltoimage --format ' + image_format + 
                           ' --quality ' + str(quality) + 
                           ' --zoom ' + str(zoom) + ' ' + 
                            html_filename + ' ' + image_filename)


"""
The logger module contains a function to initialize the logger.  Logger
warnings are printed to the monitoring report.
"""
import logging
from os.path import abspath, dirname, join
#logging.getLogger('pecos').addHandler(logging.NullHandler())

def initialize():
    """
    Initialize the pecos logger.  
    Warnings are printed to the monitoring report.
    """
    pecos_logger = logging.getLogger('pecos')
    pecos_logger.setLevel(logging.DEBUG)
    
    fh = logging.FileHandler(join(dirname(abspath(__file__)),'logfile'), mode='w')
    fh.setLevel(logging.WARNING)
    
    # all info is sent to the screen
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    
    formatter = logging.Formatter('%(message)s')
    fh.setFormatter(formatter)
    ch.setFormatter(formatter)
    if not len(pecos_logger.handlers):
        pecos_logger.addHandler(fh)
        pecos_logger.addHandler(ch)
    
    



from setuptools import setup, find_packages
from distutils.core import Extension

DISTNAME = 'pecos'
VERSION = '0.1.7'
PACKAGES = ['pecos']
EXTENSIONS = []
DESCRIPTION = 'Python package for performance monitoring of time series data.'
LONG_DESCRIPTION = open('README.md').read()
AUTHOR = 'Pecos Developers'
MAINTAINER_EMAIL = 'kaklise@sandia.gov'
LICENSE = 'Revised BSD'
URL = 'https://github.com/sandialabs/pecos'

setuptools_kwargs = {
    'zip_safe': False,
    'install_requires': ['numpy >= 1.10.4',
                         'pandas >= 0.18.0',
                         'matplotlib',
                         'jinja2'],
    'scripts': [],
    'include_package_data': True
}

setup(name=DISTNAME,
      version=VERSION,
      packages=PACKAGES,
      ext_modules=EXTENSIONS,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      author=AUTHOR,
      maintainer_email=MAINTAINER_EMAIL,
      license=LICENSE,
      url=URL,
      **setuptools_kwargs)



"""
The metrics module contains performance metrics to track system health.
"""
import pandas as pd
import numpy as np
import datetime
import logging

logger = logging.getLogger(__name__)

def qci(mask, tfilter=None, per_day=True):
    """
    Compute the quality control index (QCI), defined as:
    
    :math:`QCI=\dfrac{\sum_{d\in D}\sum_{t\in T}X_{dt}}{|DT|}`
    
    where 
    :math:`D` is the set of data columns and 
    :math:`T` is the set of timestamps in the analysis.  
    :math:`X_{dt}` is a data point for column :math:`d` time t` that passed all quality control test.  
    :math:`|DT|` is the number of data points in the analysis.
    
    Parameters
    ----------
    mask : pandas DataFrame
        Test results mask, returned from pm.get_test_results_mask()
    
    tfilter : pandas Series (optional)
        Time filter containing boolean values for each time index
        
    per_day : boolean (optional)
        Flag indicating if the results should be computed per day, default = True
    
    Returns
    -------
    pandas DataFrame with quality control index
    """
    logger.info("Compute QCI")
    
    # Remove time filter
    if tfilter is not None:
        mask = mask[tfilter]
    
    if per_day:     
        dates = [mask.index[0].date() + datetime.timedelta(days=x) for x in range(0, (mask.index[-1].date()-mask.index[0].date()).days+1)]
    
        QCI = pd.DataFrame(index=pd.to_datetime(dates))
    
        for date in dates:
            mask_date = mask.loc[date.strftime('%m/%d/%Y')]
            try:
                QCIndex = mask_date.sum().sum()/float(mask_date.shape[0]*mask_date.shape[1])
                if np.isnan(QCIndex):
                    QCIndex = 0
                QCI.loc[date, 'Quality Control Index'] = QCIndex
            except:
                QCI.loc[date, 'Quality Control Index'] = 0
    else:
        QCI = mask.sum().sum()/float(mask.shape[0]*mask.shape[1])
        
    return QCI   

def rmse(x1, x2, tfilter=None, per_day=True):
    """
    Compute the root mean squared error (RMSE), defined as:
    
    :math:`RMSE=\sqrt{\dfrac{\sum{(x_1-x_2)^2}}{n}}`
    
    where
    :math:`x_1` is a time series,
    :math:`x_2` is a time series, and 
    :math:`n` is a number of data points.
    
    Parameters
    -----------
    x1 : pandas DataFrame with a single column or pandas Series
        Data
        
    x2 : pandas DataFrame with a single column or pandas Series
        Data
         
    tfilter : pandas Series (optional)
        Time filter containing boolean values for each time index
        
    per_day : boolean (optional)
        Flag indicating if the results should be computed per day, default = True
    
    Returns
    -------
    pandas DataFrame with root mean squared error
    """
    logger.info("Compute RMSE")
    
    if type(x1) is pd.core.frame.DataFrame:
        x1 = pd.Series(x1.values[:,0], index=x1.index)
    if type(x2) is pd.core.frame.DataFrame:
        x2 = pd.Series(x2.values[:,0], index=x2.index)
        
    if tfilter is not None:
        x1 = x1[tfilter]
        x2 = x2[tfilter]
        
    def compute_rmse(data1, data2):
        val = np.sqrt(np.mean(np.power(data1 - data2,2)))
        return val
        
    if per_day:
        dates = [x1.index[0].date() + datetime.timedelta(days=x) for x in range(0, (x1.index[-1].date()-x1.index[0].date()).days+1)]
        rmse = pd.DataFrame(index=pd.to_datetime(dates))
        for date in dates:
            x1_date1 = x1.loc[date.strftime('%m/%d/%Y')] 
            x2_date2 = x2.loc[date.strftime('%m/%d/%Y')] 
            val = compute_rmse(x1_date1, x2_date2)
            rmse.loc[date, 'RMSE'] = val
    else:
        val = compute_rmse(x1, x2)
        rmse = pd.DataFrame(val, index=[0], columns=['RMSE'])

    return rmse
    
def time_integral(data, tfilter=None, per_day=True):
    """
    Compute the time integral (F) of each column in the DataFrame, defined as:
    
    :math:`F=\int{fdt}`
    
    where 
    :math:`f` is a column of data 
    :math:`dt` is the time step between observations.
    The time integral is computed using the trapezoidal rule from numpy.trapz.
    Results are given in [original data units]*seconds.
    NaN values are set to 0 for integration.
    
    Parameters
    -----------
    data : pandas DataFrame
        Data
         
    tfilter : pandas Series (optional)
        Time filter containing boolean values for each time index
        
    per_day : boolean (doptional)
        Flag indicating if the results should be computed per day, default = True
    
    Returns
    -------
    pandas DataFrame with time integral of the data, each column is named 
    'Time integral of ' + original column name
    """
    logger.info("Compute time integral")
    
    if tfilter is not None:
        data = data[tfilter]
    
    data = data.fillna(0) # fill NaN with 0
    
    def compute_integral(d):
        val = {}
        tdelta = ((d.index - d.index[0]).values)/1000000000 # convert ns to seconds
        for col in d.columns:
            val['Time integral of ' + col] = float(np.trapz(d.loc[:,col], tdelta))
        return val
        
    if per_day:
        dates = [data.index[0].date() + datetime.timedelta(days=x) for x in range(0, (data.index[-1].date()-data.index[0].date()).days+1)]
        F = pd.DataFrame(index=pd.to_datetime(dates))
        for date in dates:
            df_date = data.loc[date.strftime('%m/%d/%Y')] 
            df_int = compute_integral(df_date)
            for col in df_int:
                F.loc[date, col] = df_int[col]
    else:
        F = compute_integral(data)
        F = pd.DataFrame(F, index=[0])
    
    return F

def probability_of_detection(observed, actual, tfilter=None):
    """ 
    Compute probability of detection (PD), defined as:
        
    :math:`PD=\dfrac{TP}{TP+FN}`
    
    where 
    :math:`TP` is number of true positives and  
    :math:`FN` is the number of false negatives.
    
    Parameters
    ----------
    observed : pandas DataFrame
        Estimated conditions (True = background, False = anomolous), 
        returned from pm.get_test_results_mask()
    
    actual : pandas DataFrame
        Actual conditions, (True = background, False = anomolous)
   
    Returns
    -------
    Probability of detection, float
    """
    # Remove time filter
    if tfilter is not None:
        observed = observed[tfilter]
        actual = actual[tfilter]

    # True positive (TP) = anomolous condition where tests fail
    # TP is 1 where the statement is true, Nan where the statement is false
    TP = (observed.where(observed == False)+1) == (actual.where(actual == False)+1)
    TP_count = TP.sum().sum()
    
    # False negative (FN) = anomolous condition where tests pass
    # FN is 1 where the statement is true, Nan where the statement is false
    FN = (observed.where(observed == True)) == (actual.where(actual == False)+1)
    FN_count = FN.sum().sum()
    
    # Probability of detection
    PD = TP_count/float(TP_count+FN_count)
    
    return PD

def false_alarm_rate(observed, actual, tfilter=None):
    """ 
    Compute false alarm rate (FAR), defined as:
        
    :math:`FAR=\dfrac{TN}{TN+FP}`
    
    where 
    :math:`TN` is number of true negatives and  
    :math:`FP` is the number of false positives.
    
    Parameters
    ----------
    estimated : pandas DataFrame
        Estimated conditions (True = background, False = anomolous), 
        returned from pm.get_test_results_mask()
    
    actual : pandas DataFrame
        Actual conditions, (True = background, False = anomolous)
    
    Returns
    -------
    False alarm rate, float
    """
    # Remove time filter
    if tfilter is not None:
        observed = observed[tfilter]
        actual = actual[tfilter]

    # True negative (TN) = normal condition where tests pass
    # TN is 1 where the statement is true, Nan where the statement is false
    TN = (observed.where(observed == True)) == (actual.where(actual == True))
    TN_count = TN.sum().sum()
    
    # False positive (FP) = normal condition where tests fail
    # FP is 1 where the statement is true, Nan where the statement is false
    FP = (observed.where(observed == False)+1) == (actual.where(actual == True))
    FP_count = FP.sum().sum()
    
    # False alarm rate
    FAR = 1-TN_count/float(TN_count+FP_count)
    
    return FAR